# -*- coding: utf-8 -*-
import six

from django.core import exceptions as django_exceptions
from django.db import models
from odin import exceptions as odin_exceptions
from odin.codecs import json_codec
from odin.utils import getmeta

from baldr import form_fields

# Treat an empty JSON object as None.
EMPTY_VALUES = (None, '', {}, '{}')


class ResourceFieldDescriptor(object):
    """
    Descriptor for use with a resource field.
    """
    def __init__(self, field):
        self.field = field

    def __get__(self, instance, owner):
        if instance is None:
            raise AttributeError(
                "The '%s' attribute can only be accessed from %s instances."
                % (self.field.name, owner.__name__))

        resource = instance.__dict__[self.field.name]

        if resource in EMPTY_VALUES:
            return

        if isinstance(resource, six.string_types):
            try:
                resource = self.field.codec.loads(resource, self.field.resource_type, full_clean=False)
            except (odin_exceptions.ValidationError, odin_exceptions.CodecDecodeError):
                pass
            else:
                instance.__dict__[self.field.name] = resource

        return resource

    def __set__(self, instance, value):
        instance.__dict__[self.field.name] = value


class ResourceField(models.TextField):
    """
    Improved resource field to handle serializes/de-serializes of an Odin resource.

    This improved field is lazy in that it will not attempt to de-serialise until needed.

    This new field is also compatible with Django 1.7 migrations.
    """
    form_class = form_fields.ResourceField

    def __init__(self, resource_type, verbose_name=None, name=None, allow_subclasses=True, *args, **kwargs):
        super(ResourceField, self).__init__(verbose_name, name, *args, **kwargs)
        self.resource_type = resource_type
        self.allow_subclasses = allow_subclasses
        # This is fixed for now, this is a limitation of the current odin codecs, a refactor is needed to provide
        # codec classes.
        self.codec = json_codec
        self.codec_kwargs = dict(sort_keys=True)

    def to_python(self, value):
        if value in EMPTY_VALUES:
            return

        if isinstance(value, self.resource_type):
            return value

        if isinstance(value, six.string_types):
            try:
                return self.codec.loads(value, self.resource_type, full_clean=False)
            except odin_exceptions.CodecDecodeError as cde:
                raise django_exceptions.ValidationError(str(cde))

        raise django_exceptions.ValidationError(
            'Value provide is not a valid %s resource' % getmeta(self.resource_type).resource_name)

    def validate(self, value, model_instance):
        if not self.editable:
            # Skip validation for non-editable fields.
            return

        if value is None and not self.null:
            raise django_exceptions.ValidationError(self.error_messages['null'], code='null')

        if not self.blank and value in self.empty_values:
            raise django_exceptions.ValidationError(self.error_messages['blank'], code='blank')

        if value.__class__ is self.resource_type or (self.allow_subclasses and isinstance(value, self.resource_type)):
            try:
                value.full_clean()
            except odin_exceptions.ValidationError as ve:
                raise django_exceptions.ValidationError(str(ve.message_dict))

        else:
            raise django_exceptions.ValidationError(
                'Value provide is not a valid %s resource' % getmeta(self.resource_type).resource_name)

    def get_db_prep_save(self, value, connection):
        # Convert our JSON object to a string before we save
        if value is None:
            value = None if self.null else ""
        else:
            value = self.codec.dumps(value, **self.codec_kwargs)
        return super(ResourceField, self).get_db_prep_save(value, connection=connection)

    def contribute_to_class(self, cls, name):
        super(ResourceField, self).contribute_to_class(cls, name)
        setattr(cls, self.name, ResourceFieldDescriptor(self))

    def deconstruct(self):
        name, path, args, kwargs = super(ResourceField, self).deconstruct()
        kwargs['resource_type'] = self.resource_type
        # kwargs['codec'] = self.codec
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        defaults = {
            'form_class': self.form_class,
            'resource_type': self.resource_type
        }
        defaults.update(kwargs)
        return super(ResourceField, self).formfield(**defaults)


class ResourceListField(ResourceField):
    """
    Field that serializes/de-serializes a list of Odin resource to the database seamlessly.
    """
    form_class = form_fields.ResourceListField

    def to_python(self, value):
        if value in [None, '', {}, '{}']:  # Treat an empty JSON object as null.
            return

        if isinstance(value, (list, tuple)):
            return value

        if isinstance(value, six.string_types):
            try:
                return self.codec.loads(value, self.resource_type, full_clean=False)
            except odin_exceptions.ValidationError as ve:
                raise django_exceptions.ValidationError(str(ve.message_dict))
            except ValueError as ve:
                raise django_exceptions.ValidationError(ve.message)

        raise django_exceptions.ValidationError(
            'Value provide is not a valid %s resource' % getmeta(self.resource_type).resource_name)

    def validate(self, value, model_instance):
        if not self.editable:
            # Skip validation for non-editable fields.
            return

        if value is None and not self.null:
            raise django_exceptions.ValidationError(self.error_messages['null'], code='null')

        if not self.blank and value in self.empty_values:
            raise django_exceptions.ValidationError(self.error_messages['blank'], code='blank')

        if isinstance(value, (list, tuple)):
            errors = {}
            for idx, resource in enumerate(value):
                try:
                    super(ResourceField, self).validate(value, model_instance)
                except django_exceptions.ValidationError as ve:
                        errors[idx] = ve.message_dict
            if errors:
                raise django_exceptions.ValidationError(errors)

        else:
            raise django_exceptions.ValidationError(
                'Value provide is not a valid %s resource' % getmeta(self.resource_type).resource_name)



from django.core.exceptions import ValidationError
from django.db import models
import unittest
import odin
from baldr import model_fields


class SimpleResource(odin.Resource):
    class Meta:
        namespace = 'baldr.tests'
    name = odin.StringField()


class SimpleModel(models.Model):
    simple = model_fields.ResourceField(resource_type=SimpleResource)
    simple_list = model_fields.ResourceListField(resource_type=SimpleResource)


class ResourceFieldDescriptorTestCase(unittest.TestCase):
    def test_valid_values(self):
        target = SimpleModel(simple=SimpleResource(name="foo"),
                             simple_list='[{"$":"baldr.tests.SimpleResource", "name": "bar"}]')
        self.assertEqual("foo", target.simple.name)
        self.assertEqual("bar", target.simple_list[0].name)

    def test_invalid_values(self):
        target = SimpleModel(simple='{"$":"baldr.tests.SimpleResource}',
                             simple_list=123)
        self.assertEqual('{"$":"baldr.tests.SimpleResource}', target.simple)
        self.assertEqual(123, target.simple_list)

    def test_empty_field(self):
        target = SimpleModel(simple=None, simple_list='')
        self.assertIsNone(target.simple)
        self.assertIsNone(target.simple_list)

        target = SimpleModel(simple={}, simple_list='{}')
        self.assertIsNone(target.simple)
        self.assertIsNone(target.simple_list)


class ResourceFieldTestCase(unittest.TestCase):
    def test_to_python(self):
        target = model_fields.ResourceField(resource_type=SimpleResource)

        self.assertEqual(None, target.to_python(None))
        self.assertEqual(None, target.to_python(''))
        self.assertEqual(None, target.to_python({}))
        self.assertEqual(None, target.to_python('{}'))
        resource = SimpleResource(name='foo')
        self.assertEqual('foo', target.to_python(resource).name)
        self.assertEqual('bar', target.to_python('{"$":"baldr.tests.SimpleResource", "name": "bar"}').name)
        self.assertRaises(ValidationError, target.to_python, 123)



# coding=utf-8
from django.conf import settings


def accepts_header():
    """
    Resolve content type from the accepts header of a request.
    """
    def inner(request):
        return request.META.get('accepts')
    return inner


def content_type_header():
    """
    Resolve content type from the content_type header of a request.
    """
    def inner(request):
        return request.META.get('content-type')
    return inner


class DefaultString(str):
    is_default = True


def specific_default(content_type):
    """
    Specify a specific default content type.

    :param content_type: The content type to use.
    """
    def inner(_):
        return DefaultString(content_type)
    return inner


def settings_default(content_type='application/json'):
    """
    Default from ``settings.BALDR_DEFAULT_CONTENT_TYPE``.

    :param content_type: The content type to use as a fallback if setting is not defined.
    """
    def inner(_):
        return DefaultString(getattr(settings, 'BALDR_DEFAULT_CONTENT_TYPE', content_type))
    return inner



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import six

from baldr.resources import Listing

from . import constants

__all__ = (
    # Basic routes
    'route', 'action', 'collection', 'collection_action',
    'detail_route', 'detail_action',
    # Handlers
    'list_response',
    # Shortcuts
    'listing', 'create', 'detail', 'update', 'patch', 'delete'
)

_route_count = 0


# Route definition decorators

def route(func=None, name=None, path_type=constants.PATH_TYPE_COLLECTION, method=constants.GET, resource=None):
    """
    Decorator for defining an API route. Usually one of the helpers (listing,
    create, detail, update, delete) would be used in place of the route
    decorator.

    Usage::

        class ItemApi(ResourceApi):
            resource = Item

            @route(path_type=PATH_TYPE_LIST, method=GET)
            def list_items(self, request):

                ...

                return items


    :param func: Function we are routing
    :param name: Action name
    :param path_type: Type of path, list/detail or custom.
    :param method: HTTP method(s) this function responses to.
    :param resource: Specify the resource that this function encodes/decodes,
        default is the one specified on the ResourceAPI instance.

    """
    if isinstance(method, six.string_types):
        method = (method,)

    global _route_count
    route_number = _route_count
    _route_count += 1

    def inner(f):
        f.route = (route_number, path_type, method, name)
        f.resource = resource
        return f

    return inner(func) if func else inner


collection = collection_action = action = route


def resource_route(func=None, name=None, method=constants.GET, resource=None):
    return route(func, name, constants.PATH_TYPE_RESOURCE, method, resource)


detail_route = detail_action = resource_route


# Handlers

def list_response(func=None, default_offset=0, default_limit=50):
    """
    Handle processing a list. It is assumed decorator will operate on a class.
    """
    def inner(f):
        def wrapper(self, request, *args, **kwargs):
            # Get paging args from query string
            offset = kwargs['offset'] = int(request.GET.get('offset', default_offset))
            limit = kwargs['limit'] = int(request.GET.get('limit', default_limit))
            result = f(self, request, *args, **kwargs)
            if result is not None:
                if isinstance(result, tuple) and len(result) == 2:
                    result, total_count = result
                else:
                    total_count = None
                return Listing(list(result), limit, offset, total_count)
        return wrapper
    return inner(func) if func else inner


# Shortcut methods

def listing(func=None, name=None, resource=None, default_offset=0, default_limit=50):
    """
    Decorator to indicate a listing endpoint.

    :param func: Function we are routing
    :param name: Action name
    :param resource: Specify the resource that this function
        encodes/decodes, default is the one specified on the ResourceAPI
        instance.
    :param default_offset: Default value for the offset from the start of listing.
    :param default_limit: Default value for limiting the response size.

    """
    return route(list_response(func, default_offset, default_limit),
                 name, constants.PATH_TYPE_COLLECTION, constants.GET, resource)


def create(func=None, name=None, resource=None):
    """
    Decorator to indicate a creation endpoint.

    :param func: Function we are routing
    :param name: Action name
    :param resource: Specify the resource that this function
        encodes/decodes, default is the one specified on the ResourceAPI
        instance.

    """
    return route(func, name, constants.PATH_TYPE_COLLECTION, constants.POST, resource)


def detail(func=None, name=None, resource=None):
    """
    Decorator to indicate a detail endpoint.

    :param func: Function we are routing
    :param name: Action name
    :param resource: Specify the resource that this function
        encodes/decodes, default is the one specified on the ResourceAPI
        instance.

    """
    return route(func, name, constants.PATH_TYPE_RESOURCE, constants.GET, resource)


def update(func=None, name=None, resource=None):
    """
    Decorator to indicate an update endpoint.

    :param func: Function we are routing
    :param name: Action name
    :param resource: Specify the resource that this function
        encodes/decodes, default is the one specified on the ResourceAPI
        instance.

    """
    return route(func, name, constants.PATH_TYPE_RESOURCE, constants.PUT, resource)


def patch(func=None, name=None, resource=None):
    """
    Decorator to indicate a patch endpoint.

    :param func: Function we are routing
    :param name: Action name
    :param resource: Specify the resource that this function
        encodes/decodes, default is the one specified on the ResourceAPI
        instance.

    """
    return route(func, name, constants.PATH_TYPE_RESOURCE, constants.PATCH, resource)


def delete(func=None, name=None, resource=None):
    """
    Decorator to indicate a deletion endpoint.

    :param func: Function we are routing
    :param name: Action name
    :param resource: Specify the resource that this function
        encodes/decodes, default is the one specified on the ResourceAPI
        instance.

    """
    return route(func, name, constants.PATH_TYPE_RESOURCE, constants.DELETE, resource)



# -*- coding: utf-8 -*-

# Well known methods
GET = 'GET'
HEAD = 'HEAD'
POST = 'POST'
PUT = 'PUT'
PATCH = 'PATCH'
DELETE = 'DELETE'
OPTIONS = 'OPTIONS'

# Type of path
PATH_TYPE_BASE = 'base'
PATH_TYPE_DETAIL = 'detail'

PATH_TYPE_COLLECTION = 'collection'
PATH_TYPE_RESOURCE = 'resource'



# -*- coding: utf-8 -*-
from baldr.resources import Error


class ImmediateHttpResponse(Exception):
    """
    A response that should be returned Immediately!
    """
    def __init__(self, resource, status=200, headers=None):
        self.resource = resource
        self.status = status
        self.headers = headers


class ImmediateErrorHttpResponse(ImmediateHttpResponse):
    """
    An error response that should be returned Immediately!
    """
    def __init__(self, status, sub_status, message, developer_message=None, meta=None, headers=None):
        super(ImmediateErrorHttpResponse, self).__init__(
            Error(status, sub_status, message, developer_message, meta),
            status, headers
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import django

from optparse import OptionParser
from django.conf import settings
from django.core.management import call_command


def main():
    """
    The entry point for the script. This script is fairly basic. Here is a
    quick example of how to use it::

        django_test_runner.py [path-to-app]

    You must have Django on the PYTHONPATH prior to running this script. This
    script basically will bootstrap a Django environment for you.

    By default this script with use SQLite and an in-memory database. If you
    are using Python 2.6 it will just work out of the box for you.
    """
    parser = OptionParser()
    parser.add_option("--DATABASE_ENGINE", dest="DATABASE_ENGINE", default="django.db.backends.sqlite3")
    parser.add_option("--DATABASE_NAME", dest="DATABASE_NAME", default=":memory:")
    parser.add_option("--DATABASE_USER", dest="DATABASE_USER", default="")
    parser.add_option("--DATABASE_PASSWORD", dest="DATABASE_PASSWORD", default="")
    parser.add_option("--SITE_ID", dest="SITE_ID", type="int", default=1)

    options, args = parser.parse_args()

    # check for app in args
    try:
        app_path = args[0]
    except IndexError:
        print("You did not provide an app path.")
        raise SystemExit
    else:
        if app_path.endswith("/"):
            app_path = app_path[:-1]
        parent_dir, app_name = os.path.split(app_path)
        sys.path.insert(0, parent_dir)

    settings.configure(**{
        "DATABASES": {
            'default': {
                'ENGINE': options.DATABASE_ENGINE,
                'NAME': options.DATABASE_NAME,  # ':memory:',
                'USER': options.DATABASE_USER,
                'PASSWORD': options.DATABASE_PASSWORD,
                'HOST': '',
                'PORT': '',
            }
        },
        "MIDDLEWARE_CLASSES": [],
        # "SITE_ID": options.SITE_ID,
        "SECRET_KEY": "Test Key",
        "ROOT_URLCONF": "django_test_runner.urls",
        "TEMPLATE_LOADERS": (
            "django.template.loaders.app_directories.Loader",
        ),
        "TEMPLATE_DIRS": tuple(),
        "INSTALLED_APPS": (
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            # "django.contrib.sites",
            app_name,
        ),

        # Force test runner to be the pre django 1.6 test runner. This tool does not work with then new default in 1.6.
        # "TEST_RUNNER": "django.test.simple.DjangoTestSuiteRunner"
    })

    # This is to ensure that Django 1.7's app registry is populated prior to calling a command.
    if hasattr(django, 'setup'):
        django.setup()

    call_command("test", app_name)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
import inspect
import odin
import sys

from django.core.exceptions import ValidationError
from django.db import models
from odin import registration
from odin.fields import NOT_PROVIDED
from odin.mapping import FieldResolverBase, mapping_factory
from odin.utils import getmeta

from baldr.model_fields import ResourceField, ResourceListField

try:
    from odin.codecs import msgpack_codec
except ImportError:
    msgpack_codec = None


# Register support for Django Models and Validators

class ModelFieldResolver(FieldResolverBase):
    """
    Field resolver for Django Models
    """
    def get_field_dict(self):
        meta = getmeta(self.obj)
        return {f.attname: f for f in meta.fields}

registration.register_field_resolver(ModelFieldResolver, models.Model)


# Register a the Django ValidationError exception with Odin.
def django_validation_error_handler(exception, field, errors):
    if hasattr(exception, 'code') and exception.code in field.error_messages:
        message = field.error_messages[exception.code]
        if exception.params:
            message = message % exception.params
        errors.append(message)
    else:
        errors.extend(exception.messages)

registration.register_validation_error_handler(ValidationError, django_validation_error_handler)


# Model factories and helper methods.

class ModelResourceMixin(odin.Resource):
    """
    Mixin that adds some helper methods for working with resources generated from models.
    """
    class Meta:
        abstract = True

    model = None

    @classmethod
    def from_model(cls, model, context=None, **field_values):
        """
        Convert this resource into a specified to resource.

        A mapping must be defined for conversion between this resource and to_resource or an exception will be raised.
        """
        mapping = registration.get_mapping(cls.model, cls)
        return mapping(model, context).convert(**field_values)

    @classmethod
    def from_queryset(cls, queryset, single=False):
        """
        Use a query set and optimise fetching of data. This assumes that your resource has the sames fields as the
        model. This helper uses ``values_list`` and ``only`` methods to limit data processing and create resources
        faster.

        The method is a lot faster than from a model as it does not construct a full model.

        :param queryset:
        :param single:
        :return:

        """
        fields = getmeta(cls).field_map.keys()
        data = queryset.values(*fields)
        if single:
            return cls.create_from_dict(data[0])
        else:
            return [cls.create_from_dict(d) for d in data]

    def save(self, context=None, commit=True, ignore_fields=None):
        """
        Save this resource instance to the database.

        :param context: Context dict passed to each mapping function.
        :param commit: If commit=True, then the newly created model instance will be saved to the database.
        :param ignore_fields: Fields that should be ignored in mapping
        :return: Newly created model instance.

        """
        model = self.convert_to(self.model, context, ignore_fields)
        if commit:
            model.save()
        return model

    def update(self, instance, context=None, commit=True, lazy=True):
        """
        Update an existing model from this model.

        :param instance: Model instance to be updated.
        :param context: Context dict passed to each mapping function.
        :param commit: If True, then changes to the updated model instance will be saved to the database.
        :param lazy: If True, then instance and source resource are compared, if no changes are found no database,
            operation is performed.
        :return: Updated model instance.

        """
        assert isinstance(instance, self.model)

        mapping = registration.get_mapping(self.__class__, self.model)
        mapper = mapping(self, context)

        if not lazy or len(mapper.diff(instance)) > 0:
            mapper.update(instance)
            if commit:
                instance.save()
        return instance


def default_map(field):
    if field.default is models.NOT_PROVIDED:
        return NOT_PROVIDED
    return field.default

BASIC_ATTR_MAP = dict(
    null='null',
    choices='choices',
    default=default_map,
    use_default_if_not_provided=lambda _: True
)
MODEL_FIELD_MAP = [
    # (Model Field, Odin Field, attribute mappings {odin_attr: model_attr})
    (ResourceField, odin.DictAs, dict(resource='resource_type', null='null')),
    (ResourceListField, odin.ListOf, dict(resource='resource_type', null='null')),

    (models.AutoField, odin.StringField, dict()),
    (models.DateTimeField, odin.DateTimeField, BASIC_ATTR_MAP),
    (models.TimeField, odin.TimeField, BASIC_ATTR_MAP),
    (models.URLField, odin.UrlField, BASIC_ATTR_MAP),
    (models.IntegerField, odin.IntegerField, BASIC_ATTR_MAP),
    (models.FloatField, odin.FloatField, BASIC_ATTR_MAP),
    (models.BooleanField, odin.BooleanField, BASIC_ATTR_MAP),
    (models.CharField, odin.StringField, dict(BASIC_ATTR_MAP, max_length='max_length')),
    (models.TextField, odin.StringField, BASIC_ATTR_MAP),
]


def field_factory(model_field):
    """
    Return an equivalent odin field from a Django model field.
    :param model_field:
    :return:
    """
    for mf, of, attrs in MODEL_FIELD_MAP:
        if isinstance(model_field, mf):
            attrs = {
                oa: (ma(model_field) if callable(ma) else getattr(model_field, ma))
                for oa, ma in attrs.items()
            }
            attrs['validators'] = getattr(model_field, 'validators')
            return of(**attrs)


NO_REVERSE_FIELDS = [
    # Model Field, check method
    (models.DateField, lambda f: f.auto_now_add or f.auto_now),
]


def field_in_filters(model_field, filters):
    """
    Check if a supplied model_field matches a set of filters.

    :param model_field:
    :param filters:
    :return:

    """
    for mf, filter_ in filters:
        if isinstance(model_field, mf):
            if filter_(model_field):
                return True
    return False


def model_resource_factory(model, module=None, base_resource=odin.Resource, resource_mixins=None,
                           exclude_fields=None, include_fields=None, generate_mappings=True,
                           return_mappings=False, additional_fields=None, resource_type_name=None,
                           reverse_exclude_fields=None):
    """
    Factory method for generating a resource from a existing Django model.

    Usage::

        class Person(models.Model):
            name = models.CharField(max_length=50)
            age = models.IntegerField()

        PersonResource = model_resource_factory(Person)

    :param model: The Django model to generate resource from.
    :param module: Module you want the class to be a member of; default uses the calling module. This value can be the
        name of another module (eg the __name__ field in a module).
    :param base_resource: Base resource to extend from; default is ``odin.Resource``.
    :param resource_mixins: Any additional mixin resources; default ``baldr.models.ModelResourceMixin``.
    :param exclude_fields: Any fields that should be excluded from the resource.
    :param include_fields: Explicitly define what fields that should be included on the resource.
    :param generate_mappings: Generate mappings between resource and model (in both directions).
    :param return_mappings: Return the mappings along with the model resource (returns a
        tuple(Resource, ForwardMapping, ReverseMapping).
    :param additional_fields: Any additional fields that should be appended to the resource, these can override fields
        from the model.
    :param resource_type_name: Name of the resource created by the factory (default is the name of the model)
    :param reverse_exclude_fields: Excluded fields from reverse mapping.

    """
    resource_mixins = resource_mixins or []
    bases = tuple(resource_mixins + [ModelResourceMixin, base_resource])
    attrs = {}
    model_opts = model._meta
    resource_type_name = resource_type_name or model_opts.object_name

    # Determine the calling module
    if module is None:
        frame = inspect.stack()[1]
        module = inspect.getmodule(frame[0])

    # Append fields
    exclude_fields = exclude_fields or []
    reverse_exclude_fields = reverse_exclude_fields or []
    for mf in model_opts.fields:
        if mf.attname in exclude_fields:
            continue

        # Create an odin version of the field.
        field = field_factory(mf)
        if field:
            attrs[mf.attname] = field

        # Check if the field should not be reversed
        if field_in_filters(mf, NO_REVERSE_FIELDS):
            reverse_exclude_fields.append(mf.attname)

    # Add any additional fields.
    if additional_fields:
        assert isinstance(additional_fields, dict)
        for attname, field in additional_fields.items():
            attrs[attname] = field

    # Setup other require attributes and create type
    if isinstance(module, str):
        module = sys.modules[module]
    attrs['__module__'] = module.__name__
    attrs['model'] = model
    resource_type = type(resource_type_name, bases, attrs)

    # Generate mappings
    forward_mapping, reverse_mapping = None, None
    if generate_mappings:
        forward_mapping, reverse_mapping = mapping_factory(
            model, resource_type, reverse_exclude_fields=reverse_exclude_fields
        )

    if return_mappings:
        return resource_type, forward_mapping, reverse_mapping
    else:
        return resource_type


# Register Django Promises (used by translated strings) with Odin codecs

# TODO: this is not quite correct, will be corrected in updated serialisation tools current in development
# for odin
# json_codec.JSON_TYPES[Promise] = force_text
# if msgpack_codec:
#     msgpack_codec.TYPE_SERIALIZERS[Promise] = force_text



# -*- coding: utf-8 -*-
import odin


class Listing(odin.Resource):
    """
    Response for listing results. This includes offset, count support for
    paging etc.

    """
    class Meta:
        namespace = None

    # Wrapper to provide code completion
    def __init__(self, results, limit, offset=0, total_count=None):
        super(Listing, self).__init__(results, limit, offset, total_count)

    results = odin.ArrayField(
        help_text="The list of resources."
    )
    limit = odin.IntegerField(
        help_text="The resource limit in the result set."
    )
    offset = odin.IntegerField(
        default=0,
        help_text="The offset within the result set."
    )
    total_count = odin.IntegerField(
        null=True,
        help_text="The total number of items in the result set."
    )


class Error(odin.Resource):
    """
    Response returned for errors.

    The *meta* field should be utilised to provide additional information that
    is specific to the error. Eg if validation failed then meta would contain
    an object that maps field names to error messages.

    """
    class Meta:
        namespace = None

    status = odin.IntegerField(
        help_text="HTTP status code of the response."
    )
    sub_status = odin.IntegerField(
        help_text="Sub-status code for more fine grained error results."
    )
    message = odin.StringField(
        help_text="A message that can be used for an end user."
    )
    developer_message = odin.StringField(
        null=True,
        help_text="More complex error message suitable for a developer."
    )
    meta = odin.StringField(
        null=True,
        help_text="Additional meta information that can help to solve issues."
    )



# -*- coding: utf-8 -*-
import odin
import six

from django.core import exceptions as django_exceptions
from django.forms import widgets
from django.forms.fields import CharField
from django.utils.translation import ugettext_lazy as _
from odin import exceptions as odin_exceptions
from odin.codecs import json_codec
from odin.utils import getmeta


class ResourceField(CharField):
    """
    Form field that wraps an Odin resource.
    """

    widget = widgets.Textarea

    default_error_messages = {
        'invalid': _('This field is not a valid %s resource.'),
    }

    def __init__(self, resource_type, codec=json_codec, codec_kargs=None, *args, **kwargs):
        assert issubclass(resource_type, odin.Resource)
        super(ResourceField, self).__init__(*args, **kwargs)

        self.resource_type = resource_type
        self.codec = codec
        self.codec_kwargs = dict(indent=2, sort_keys=True) if codec_kargs is None else codec_kargs

    def prepare_value(self, value):
        if value is None:
            return ''
        elif isinstance(value, (list, tuple, odin.Resource)):
            return self.codec.dumps(value, **self.codec_kwargs)
        else:
            return str(value)

    def to_python(self, value):
        if value is None or value == '':
            return

        if isinstance(value, self.resource_type):
            return value

        if isinstance(value, six.string_types):
            try:
                return self.codec.loads(value, self.resource_type, full_clean=False)
            except odin_exceptions.ValidationError as ve:
                if hasattr(ve, 'message_dict'):
                    raise django_exceptions.ValidationError(str(ve.message_dict))
                else:
                    raise django_exceptions.ValidationError(ve.messages)
            except odin_exceptions.CodecDecodeError as cde:
                raise django_exceptions.ValidationError(str(cde))
            except ValueError as ve:
                raise django_exceptions.ValidationError(str(ve))

        raise django_exceptions.ValidationError(
            self.error_messages['invalid'] % getmeta(self.resource_type).resource_name, code='invalid'
        )

    def validate(self, value):
        super(ResourceField, self).validate(value)

        if value is None:
            return

        if isinstance(value, self.resource_type):
            try:
                value.full_clean()
            except odin_exceptions.ValidationError as ve:
                if hasattr(ve, 'message_dict'):
                    raise django_exceptions.ValidationError(str(ve.message_dict))
                else:
                    raise django_exceptions.ValidationError(ve.messages)
        else:
            raise django_exceptions.ValidationError(
                self.error_messages['invalid'] % getmeta(self.resource_type).resource_name, code='invalid')


class ResourceListField(ResourceField):
    """
    Form field that wraps a list of odin fields.
    """
    def to_python(self, value):
        if value is None or value == '':
            return

        if isinstance(value, (list, tuple)):
            return value

        if isinstance(value, six.string_types):
            try:
                return self.codec.loads(value, self.resource_type, full_clean=False)
            except odin_exceptions.CodecDecodeError as cde:
                raise django_exceptions.ValidationError(str(cde))
            except ValueError as ve:
                raise django_exceptions.ValidationError(str(ve))

        raise django_exceptions.ValidationError(
            self.error_messages['invalid'] % getmeta(self.resource_type).resource_name, code='invalid')

    def validate(self, value):
        if value is None:
            return

        elif isinstance(value, (list, tuple)):
            errors = {}
            for idx, resource in enumerate(value):
                try:
                    super(ResourceField, self).validate(value)
                except django_exceptions.ValidationError as ve:
                        errors[idx] = ve.message_dict
            if errors:
                raise django_exceptions.ValidationError(errors)

        # Unknown type
        else:
            raise django_exceptions.ValidationError(
                self.error_messages['invalid'] % getmeta(self.resource_type).resource_name, code='invalid'
            )



# -*- coding: utf-8 -*-
from baldr.api import ResourceApi
from baldr.exceptions import ImmediateErrorHttpResponse


class LoginRequiredMixin(ResourceApi):
    """
    Ensure that a user has logged in.
    """
    def handle_authorisation(self, request):
        """
        Evaluate if a request is authorised.

        :param request: The current Django request object.

        """
        if not request.user.is_authenticated():
            raise ImmediateErrorHttpResponse(403, 0, 'Login required')



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging
import six
import sys

from collections import OrderedDict
from django.conf import settings
from django.conf.urls import url, include
from django.core.exceptions import PermissionDenied
from django.http import HttpResponse, Http404
from django.views.decorators.csrf import csrf_exempt

from odin.codecs import json_codec
from odin.exceptions import CodecDecodeError, ValidationError
from odin.utils import getmeta, lazy_property

from .. import content_type_resolvers
from ..resources import Error
from ..exceptions import ImmediateErrorHttpResponse, ImmediateHttpResponse
from .constants import *  # noqa
from .route_decorators import *  # noqa


logger = logging.getLogger(__name__)

CODECS = {json_codec.CONTENT_TYPE: json_codec}
# Attempt to load other codecs that have dependencies
try:
    from odin.codecs import msgpack_codec
except ImportError:
    pass
else:
    CODECS[msgpack_codec.CONTENT_TYPE] = msgpack_codec


class ResourceApiCommon(object):
    # The resource this API is modelled on.
    resource = None
    resource_id_regex = r'\d+'

    # Handlers used to resolve the request content-type of the request body.
    # These are checked in the order defined until one returns a content-type.
    request_type_resolvers = [
        content_type_resolvers.content_type_header(),
        content_type_resolvers.accepts_header(),
        content_type_resolvers.settings_default(),
    ]

    # Handlers used to resolve the response content-type.
    # These are checked in the order defined until one returns a content-type.
    response_type_resolvers = [
        content_type_resolvers.accepts_header(),
        content_type_resolvers.content_type_header(),
        content_type_resolvers.settings_default(),
    ]

    # Codecs that are supported for Encoding/Decoding resources.
    registered_codecs = CODECS
    url_prefix = r''

    def __init__(self, api_name=None):
        if api_name:
            self.api_name = api_name
        elif not hasattr(self, 'api_name'):
            self.api_name = "%ss" % getmeta(self.resource).name

    def url(self, regex, view, kwargs=None, name=None, prefix=''):
        """
        Behaves like the django built in ``url`` method but constrains the URL to the API name.

        Regex should be a part regex that applies only to the targeted method ie::

            self.url("(\d+)", ...)

        """
        if regex:
            return url(r'^%s/%s/?$' % (self.url_prefix + self.api_name.lower(), regex), view, kwargs, name, prefix)
        else:
            return url(r'^%s/?$' % (self.url_prefix + self.api_name.lower()), view, kwargs, name, prefix)

    @property
    def urls(self):
        """
        Return url conf for resource object.
        """
        return self.base_urls()

    def resolve_request_type(self, request):
        """
        Resolve the request content type from the request.

        :returns: Identified content type; or ``None`` if content type is not identified.

        """
        for resolver in self.request_type_resolvers:
            content_type = resolver(request)
            if content_type:
                return content_type

    def resolve_response_type(self, request):
        """
        Resolve the response content type from the request.

        :returns: Identified content type; or ``None`` if content type is not identified.

        """
        for resolver in self.response_type_resolvers:
            content_type = resolver(request)
            if content_type:
                return content_type

    @staticmethod
    def handle_500(request, exception):
        """
        Handle *un-handled* exceptions

        :param request: The request object.
        :param exception: The exception that was un-handled.
        :return: An ``HttpError`` response resource.

        """
        exc_info = sys.exc_info()

        # This is an unknown exception, return an unknown error message.
        if settings.DEBUG:
            # If we are in debug mode return more details and the stack-trace.
            import traceback
            the_trace = '\n'.join(traceback.format_exception(*exc_info))
            return Error(500, 50000, "An unknown error has occurred, the developers have been notified.",
                         str(exception), the_trace)
        else:
            logger.error('Internal Server Error: %s', request.path, exc_info=exc_info, extra={
                'status_code': 500,
                'request': request
            })
            return Error(500, 50000, "An unknown error has occurred, the developers have been notified.")

    def base_urls(self):
        """
        Base URL mappings for this API.
        """
        return []

    @staticmethod
    def decode_body(request):
        """
        Helper method that ensures that decodes any body content into a string object (this is needed by the json
        module for example).
        """
        body = request.body
        if isinstance(body, bytes):
            return body.decode('UTF8')
        return body

    def resource_from_body(self, request, allow_multiple=False, resource=None):
        """
        Get a resource instance from ``request.body``.
        """
        resource = resource or self.resource

        try:
            body = self.decode_body(request)
        except UnicodeDecodeError as ude:
            raise ImmediateErrorHttpResponse(400, 40099, "Unable to decode request body.", str(ude))

        try:
            resource = request.request_codec.loads(body, resource=resource, full_clean=False)
        except ValueError as ve:
            raise ImmediateErrorHttpResponse(400, 40098, "Unable to load resource.", str(ve))
        except CodecDecodeError as cde:
            raise ImmediateErrorHttpResponse(400, 40096, "Unable to decode body.", str(cde))

        # Check an array of data hasn't been supplied
        if not allow_multiple and isinstance(resource, list):
            raise ImmediateErrorHttpResponse(400, 40097, "Expected a single resource not a list.")

        return resource

    def dispatch_to_view(self, view, request, *args, **kwargs):
        raise NotImplementedError()

    def wrap_view(self, view):
        """
        This method provides the main entry point for URL mappings in the ``base_urls`` method.
        """
        @csrf_exempt
        def wrapper(request, *args, **kwargs):
            # Resolve content type used to encode/decode request/response content.
            response_type = self.resolve_response_type(request)
            request_type = self.resolve_request_type(request)
            try:
                request.request_codec = self.registered_codecs[request_type]
                request.response_codec = response_codec = self.registered_codecs[response_type]
            except KeyError:
                # This is just a plain HTTP response, we can't provide a rich response when the content type is unknown
                return HttpResponse(content="Content cannot be returned in the format requested.", status=406)

            try:
                result = self.dispatch_to_view(view, request, *args, **kwargs)
            except Http404 as e:
                # Item is not found.
                status = 404
                resource = Error(status, 40400, str(e))

            except ImmediateHttpResponse as e:
                # An exception used to return a response immediately, skipping any further processing.
                response = HttpResponse(
                    response_codec.dumps(e.resource),
                    content_type=response_codec.CONTENT_TYPE,
                    status=e.status
                )
                for key, value in (e.headers or {}).items():
                    response[key] = value
                return response

            except ValidationError as e:
                # Validation of a resource has failed.
                status = 400
                if hasattr(e, 'message_dict'):
                    resource = Error(status, 40000, "Fields failed validation.", meta=e.message_dict)
                else:
                    resource = Error(status, 40000, str(e))

            except PermissionDenied as e:
                status = 403
                resource = Error(status, 40300, "Permission denied", str(e))

            except NotImplementedError:
                # A mixin method has not been implemented, as defining a mixing is explicit this is considered a server
                # error that should be addressed.
                status = 501
                resource = Error(status, 50100, "This method has not been implemented.")

            except Exception as e:
                # Special case when a request raises a 500 error. If we are in debug mode and a default is used (ie
                # request does not explicitly specify a content type) fall back to the Django default exception page.
                if settings.DEBUG and getattr(response_type, 'is_default', False):
                    raise
                # Catch any other exceptions and pass them to the 500 handler for evaluation.
                resource = self.handle_500(request, e)
                status = resource.status

            else:
                if isinstance(result, tuple) and len(result) == 2:
                    resource, status = result
                else:
                    resource = result
                    status = 204 if result is None else 200  # Return 204 (No Content) if result is None.

            if resource is None:
                return HttpResponse(status=status)
            elif isinstance(resource, HttpResponse):
                return resource
            else:
                return HttpResponse(
                    response_codec.dumps(resource),
                    content_type=response_codec.CONTENT_TYPE,
                    status=status
                )
        return wrapper


class ResourceApiBase(type):
    def __new__(mcs, name, bases, attrs):
        super_new = super(ResourceApiBase, mcs).__new__

        # attrs will never be empty for classes declared in the standard way
        # (ie. with the `class` keyword). This is quite robust.
        if name == 'NewBase' and attrs == {}:
            return super_new(mcs, name, bases, attrs)

        parents = [
            b for b in bases
            if isinstance(b, ResourceApiBase) and not (b.__name__ == 'NewBase' and b.__mro__ == (b, object))
        ]
        if not parents:
            # If this isn't a subclass of don't do anything special.
            return super_new(mcs, name, bases, attrs)

        routes = []

        # Get local routes and sort them by route number
        for view, obj in attrs.items():
            if callable(obj) and hasattr(obj, 'route'):
                route_number, path_type, method, action_name = obj.route
                routes.append((route_number, path_type, method, action_name, view))
                delattr(obj, 'route')
        routes = sorted(routes, key=lambda o: o[0])

        # Get routes from parent objects
        for parent in parents:
            if hasattr(parent, 'routes'):
                routes.extend(parent.routes)

        attrs['routes'] = routes

        return super_new(mcs, name, bases, attrs)


@six.add_metaclass(ResourceApiBase)
class ResourceApi(ResourceApiCommon):
    """
    Base class for version 2 of the Resource API.

    The API now uses a meta class to extract routes from methods that have
    been decorated with route information.

    """
    # Table containing lookup information to simplify dispatch of incoming
    # requests to the appropriate views
    route_table = None

    # Respond to the options method.
    respond_to_options = True

    def base_urls(self):
        url_table = OrderedDict()
        route_table = {}
        for route_ in self.routes:
            route_number, path_type, methods, action_name, view = route_
            route_key = "%s-%s" % (path_type, action_name) if action_name else path_type

            # Populate url_table
            if route_key not in url_table:
                if path_type == PATH_TYPE_COLLECTION:
                    regex = action_name or r''
                else:
                    if action_name:
                        regex = r'(?P<resource_id>%s)/%s' % (self.resource_id_regex, action_name)
                    else:
                        regex = r'(?P<resource_id>%s)' % self.resource_id_regex

                url_table[route_key] = self.url(regex, self.wrap_view(route_key))

            # Populate route table
            method_map = route_table.setdefault(route_key, {})
            for method in methods:
                method_map[method] = view

            # Add options
            if self.respond_to_options:
                method_map.setdefault(constants.OPTIONS, 'options_response')

        # Store the route table at the same time
        self.route_table = dict(route_table)
        return list(url_table.values())

    def dispatch_to_view(self, route_key, request, *args, **kwargs):
        """
        Primary method used to dispatch incoming requests to the appropriate method.
        """
        routes = self.route_table[route_key]
        try:
            request_method = routes[request.method]
        except KeyError:
            allow = ','.join(routes.keys())
            raise ImmediateErrorHttpResponse(405, 40500, "Method not allowed", headers={'Allow': allow},
                                             meta={'allow': allow})
        else:
            method = getattr(self, request_method, None)

        # Authorisation hook
        if hasattr(self, 'handle_authorisation'):
            self.handle_authorisation(request)

        # Allow for a pre_dispatch hook, a response from pre_dispatch would indicate an override of kwargs
        if hasattr(self, 'pre_dispatch'):
            response = self.pre_dispatch(request, **kwargs)
            if response is not None:
                kwargs = response

        # Apply route key to kwargs on OPTIONS requests
        if request.method == constants.OPTIONS:
            kwargs['route_key'] = route_key

        result = method(request, **kwargs)

        # Allow for a post_dispatch hook, the response of which is returned
        if hasattr(self, 'post_dispatch'):
            return self.post_dispatch(request, result)
        else:
            return result

    def options_response(self, request, route_key, **kwargs):
        routes = self.route_table[route_key]
        response = HttpResponse(status=204)
        response['Allow'] = ','.join(method for method, route in routes.items() if route != 'options_response')
        return response


class ApiCollection(object):
    """
    Collection of several resource API's.
    Along with helper methods for building URL patterns.
    ::
        urlpatterns += Api(
            ApiCollection(
                UserApi(),
                MyApi(),
            )
        ).patterns()
    """
    def __init__(self, *resource_apis, **kwargs):
        self.api_name = kwargs.pop('api_name', 'api')
        self.resource_apis = resource_apis

    @lazy_property
    def urls(self):
        urls = []
        for resource_api in self.resource_apis:
            urls.extend(resource_api.urls)
        return urls

    def include(self, namespace=None):
        return include(self.urls, namespace)

    def patterns(self, api_name=None):
        api_name = api_name or self.api_name
        return [url(r'^%s/' % api_name, self.include())]


class ApiVersion(ApiCollection):
    """
    A versioned collection of several resource API's.
    Along with helper methods for building URL patterns.
    """
    def __init__(self, *resource_apis, **kwargs):
        kwargs.setdefault('api_name', kwargs.pop('version', 'v1'))
        super(ApiVersion, self).__init__(*resource_apis, **kwargs)


class Api(object):
    """
    An API (made up of versions).
    ::
        urlpatterns += Api(
            ApiVersion(
                UserApi(),
                MyApi(),
                version='v1',
            )
        ).patterns()
    """
    def __init__(self, *versions, **kwargs):
        self.versions = versions
        self.api_name = kwargs.get('api_name', 'api')

    def patterns(self):
        urls = [url(r'^%s/%s/' % (self.api_name, v.api_name), v.include()) for v in self.versions]
        urls.append(url(r'^%s/$' % self.api_name, self._unknown_version))
        return urls

    def _unknown_version(self, _):
        supported_versions = [v.api_name for v in self.versions]
        return HttpResponse(
            "Unsupported API version. Available versions: %s" % ', '.join(supported_versions),
            status=418  # I'm a teapot... Is technically a bad request but makes sense to have a different status code.
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import

__version__ = "0.8.dev0"
default_app_config = 'baldr.app.BaldrAppConfig'


def _ensure_registration():
    """
    Ensure that model type resolvers and validation tools are registered with Odin.

    This provides support for Django Models and use of Django Validators in Odin resources.
    """
    from . import models  # NoQA
_ensure_registration()



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from django.shortcuts import get_object_or_404
from odin import registration
from odin.exceptions import CodecDecodeError

from . import ResourceApi, listing, collection, create, detail, update, patch, delete
from ..exceptions import ImmediateErrorHttpResponse


class ModelResourceApi(ResourceApi):
    """
    Provides an API for working with Django models.

    This API assumes that mappings have been defined between between the model and a suitable resource.

    """
    # Model this API deals with
    model = None
    # Model field to use for single model queries.
    model_id_field = 'pk'
    # Mapping to use for mapping to model
    to_model_mapping = None
    # Mapping to use for mapping to resource
    to_resource_mapping = None

    def __init__(self, *args, **kwargs):
        super(ModelResourceApi, self).__init__(*args, **kwargs)

        assert self.model, "A model has not been provided."

        # Attempt to resolve mappings
        if self.to_model_mapping is None:
            self.to_model_mapping = registration.get_mapping(self.resource, self.model)
        if self.to_resource_mapping is None:
            self.to_resource_mapping = registration.get_mapping(self.model, self.resource)

    def get_queryset(self, request):
        return self.model.objects.all()

    def get_instance(self, request, resource_id):
        return get_object_or_404(self.get_queryset(request), **{
            self.model_id_field: resource_id
        })

    def update_instance_from_body(self, request, instance, resource=None, ignore_fields=('id', 'pk')):
        """
        Get a resource that merges an instance and the request body.
        :param request:
        :param instance:
        :param resource:
        :param ignore_fields:
        :return:

        """
        resource = resource or self.resource

        try:
            body = self.decode_body(request)
        except UnicodeDecodeError as ude:
            raise ImmediateErrorHttpResponse(400, 40100, "Unable to decode request body.", str(ude))

        try:
            resource = request.request_codec.loads(body, resource=resource, full_clean=False,
                                                   default_to_not_supplied=True)
        except ValueError as ve:
            raise ImmediateErrorHttpResponse(400, 40098, "Unable to load resource.", str(ve))
        except CodecDecodeError as cde:
            raise ImmediateErrorHttpResponse(400, 40096, "Unable to decode body.", str(cde))

        # Update only the supplied fields
        self.to_model_mapping(resource).update(instance, ignore_fields=ignore_fields)

        return resource

    def save_model(self, request, instance, is_new=False):
        instance.save()
        return instance


class CollectionMixin(ModelResourceApi):
    """
    Mixin that provides a collection response.
    """
    @collection
    def object_collection(self, request):
        queryset = self.get_queryset(request)
        return self.to_resource_mapping.apply(queryset)


class ListMixin(ModelResourceApi):
    """
    Mixin that provides a basic paged listing response.
    """
    @listing
    def object_list(self, request, limit, offset):
        queryset = self.get_queryset(request)
        results = queryset[offset:offset+limit]
        return self.to_resource_mapping.apply(results), len(queryset)


class CreateMixin(ModelResourceApi):
    """
    Mixin that provides a basic creation method.
    """
    @create
    def object_create(self, request):
        resource = self.resource_from_body(request)
        instance = self.to_model_mapping.apply(resource)
        instance.id = None
        self.save_model(request, instance, True)
        return self.to_resource_mapping.apply(instance), 201


class DetailMixin(ModelResourceApi):
    """
    Mixin that provides a basic full detail method.
    """
    @detail
    def object_detail(self, request, resource_id):
        instance = self.get_instance(request, resource_id)
        return self.to_resource_mapping.apply(instance)


class UpdateMixin(ModelResourceApi):
    """
    Mixin that provides a basic model update method.
    """
    @update
    def object_update(self, request, resource_id):
        instance = self.get_instance(request, resource_id)
        resource = self.resource_from_body(request)
        self.to_model_mapping(resource).update(instance, ignore_fields=('id', 'pk'))
        self.save_model(request, instance, False)
        return self.to_resource_mapping.apply(instance)


class PatchMixin(ModelResourceApi):
    """
    Mixin that provides a basic model update method.
    """
    @patch
    def object_update(self, request, resource_id):
        instance = self.get_instance(request, resource_id)
        self.update_instance_from_body(request, instance)
        self.save_model(request, instance, False)
        return self.to_resource_mapping.apply(instance)


class DeleteMixin(ModelResourceApi):
    """
    Mixin that provides a basic delete method.
    """
    @delete
    def object_delete(self, request, resource_id):
        self.get_instance(request, resource_id).delete()



from __future__ import absolute_import
from django import test
from django.test.client import RequestFactory
from .. import content_type_resolvers


class ContentTypeResolvers(test.TestCase):
    def setUp(self):
        self.factory = RequestFactory()

    def test_accepts_header_no_header(self):
        request = self.factory.get('/api/foo/')
        target = content_type_resolvers.accepts_header()
        actual = target(request)
        self.assertIsNone(actual)

    def test_accepts_header_json_header(self):
        request = self.factory.get('/api/foo/', accepts='application/json')
        target = content_type_resolvers.accepts_header()
        actual = target(request)
        self.assertEqual('application/json', actual)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import six
import odin

from collections import OrderedDict
from django.forms import fields
from django.forms.forms import DeclarativeFieldsMetaclass, BaseForm
from django.forms.utils import ErrorList
from odin.exceptions import ValidationError
from odin.utils import getmeta

ALL_FIELDS = '__all__'

NO_OP = lambda v: v

# Common field options
COMMON_OPTIONS = (
    ('null', lambda v: not bool(v), 'required'),
    # ('verbose_name', lambda v: v.capitalize(), 'label'),
    ('doc_text', NO_OP, 'help_text'),
)
CHOICE_OPTOINS = (
    ('choices', NO_OP, 'choices'),
)

FORM_FIELD_MAP = {
    odin.DateTimeField: (fields.DateTimeField, None),
    odin.DateField: (fields.DateField, None),
    odin.TimeField: (fields.TimeField, None),
    odin.HttpDateTimeField: (fields.DateTimeField, None),
    odin.TimeStampField: (fields.DateTimeField, None),
    odin.UrlField: (fields.URLField, None),
    odin.IntegerField: (fields.IntegerField, None),
    odin.FloatField: (fields.FloatField, None),
    odin.BooleanField: (fields.BooleanField, None),
    odin.StringField: (fields.CharField, (('max_length', NO_OP, 'max_length'),)),
}


def construct_instance(form, instance, fields=None, exclude=None):
    opts = instance._meta

    cleaned_data = form.cleaned_data
    for f in opts.fields:
        if fields is not None and f.name not in fields:
            continue
        if exclude and f.name in exclude:
            continue

        f.value_to_object(instance, cleaned_data[f.name])

    return instance


def construct_field(field, **kwargs):
    if field.choices:
        form_field = fields.ChoiceField
        field_options = CHOICE_OPTOINS
    else:
        try:
            form_field, field_options = FORM_FIELD_MAP[field.__class__]
        except KeyError:
            return

    option_values = {ff: t(getattr(field, rf)) for rf, t, ff in COMMON_OPTIONS}
    if field_options:
        option_values.update((ff, t(getattr(field, rf))) for rf, t, ff in field_options)
    option_values.update(kwargs)
    return form_field(**option_values)


def fields_for_resource(resource, fields=None, exclude=None, widgets=None,
                        resourcefield_callback=None, localized_fields=None,
                        labels=None, help_texts=None, error_messages=None):
    field_list = []
    ignored = []
    opts = getmeta(resource)

    for f in opts.all_fields:
        if fields is not None and f.name not in fields:
            continue
        if exclude and f.name in exclude:
            continue

        kwargs = {}
        if widgets and f.name in widgets:
            kwargs['widget'] = widgets[f.name]
        if localized_fields == ALL_FIELDS or (localized_fields and f.name in localized_fields):
            kwargs['localize'] = True
        if labels and f.name in labels:
            kwargs['label'] = labels[f.name]
        if help_texts and f.name in help_texts:
            kwargs['help_text'] = help_texts[f.name]
        if error_messages and f.name in error_messages:
            kwargs['error_messages'] = error_messages[f.name]

        if resourcefield_callback is None:
            formfield = construct_field(f, **kwargs)
        elif not callable(resourcefield_callback):
            raise TypeError('resourcefield_callback must be a function or callable')
        else:
            formfield = resourcefield_callback(f, **kwargs)

        if formfield:
            field_list.append((f.name, formfield))
        else:
            ignored.append(f.name)

    field_dict = OrderedDict(field_list)
    if fields:
        field_dict = OrderedDict(
            ((f, field_dict.get(f)) for f in fields
             if ((not exclude) or (exclude and f not in exclude)) and (f not in ignored))
        )
    return field_dict


class ResourceFormOptions(object):
    def __init__(self, options=None):
        self.resource = getattr(options, 'resource', None)
        self.fields = getattr(options, 'fields', None)
        self.exclude = getattr(options, 'exclude', None)
        self.widgets = getattr(options, 'widgets', None)
        self.localized_fields = getattr(options, 'localized_fields', None)
        self.labels = getattr(options, 'labels', None)
        self.help_texts = getattr(options, 'help_texts', None)
        self.error_messages = getattr(options, 'error_messages', None)


class ResourceFormMetaclass(DeclarativeFieldsMetaclass):
    """
    Metaclass that collects Fields declared on the base classes.
    """
    def __new__(mcs, name, bases, attrs):
        resourcefield_callback = attrs.pop('resourcefield_callback', None)

        new_class = super(ResourceFormMetaclass, mcs).__new__(mcs, name, bases, attrs)

        if bases == (BaseResourceForm,):
            return new_class

        opts = new_class._meta = ResourceFormOptions(getattr(new_class, 'Meta', None))

        for opt in ['fields', 'exclude']:
            value = getattr(opts, opt)
            if isinstance(value, six.string_types) and value != ALL_FIELDS:
                msg = (
                    "%(resource)s.Meta.%(opt)s cannot be a string. "
                    "Did you mean to type: ('%(value)s',)?" % {
                        'resource': new_class.__name__, 'opt': opt, 'value': value,
                    }
                )
                raise TypeError(msg)

        if opts.resource:
            if opts.fields == ALL_FIELDS:
                opts.fields = None

            fields = fields_for_resource(opts.resource, opts.fields, opts.exclude,
                                         opts.widgets, resourcefield_callback,
                                         opts.localized_fields, opts.labels,
                                         opts.help_texts, opts.error_messages)

            fields.update(new_class.declared_fields)
        else:
            fields = new_class.declared_fields

        new_class.base_fields = fields

        return new_class


class BaseResourceForm(BaseForm):
    def __init__(self, data=None, files=None, auto_id='id_%s', prefix=None,
                 initial=None, error_class=ErrorList, label_suffix=None,
                 empty_permitted=False, instance=None):
        opts = self._meta
        if opts.resource is None:
            raise ValueError('ResourceForm has no resource class specified.')
        if instance is None:
            # if we didn't get an instance, instantiate a new one
            self.instance = opts.resource()
            object_data = {}
        else:
            self.instance = instance
            object_data = instance.as_dict()
        # if initial was provided, it should override the values from instance
        if initial is not None:
            object_data.update(initial)

        super(BaseResourceForm, self).__init__(data, files, auto_id, prefix, object_data,
                                               error_class, label_suffix, empty_permitted)

    def _update_errors(self, errors):
        self.add_error(None, errors)

    def _post_clean(self):
        opts = self._meta

        self.instance = construct_instance(self, self.instance, opts.fields, opts.exclude)

        try:
            self.instance.full_clean()
        except ValidationError as e:
            self._update_errors(e)


class ResourceForm(six.with_metaclass(ResourceFormMetaclass, BaseResourceForm)):
    pass
